/************************************************************************/
/* 	Horizonator	
/*	$id$
/*	File	: 					Horizon_map_generator.cpp
/*	Version	:					1.5
/*	Descript:					        (  )
/*								         Horizon Maping
/*	Autor	:					Morock
/*	Created :					11/30/2006
/************************************************************************/


#include <stdio.h>
#include <conio.h>
#include <d3d9.h>
#include <D3dx9tex.h>
#include <Dxerr9.h>
#include <wchar.h>


#include "cParams.h"

#define TREE_HEIGHT 4.0f	//Tree height, that will add to a terrain height in plase of tree is.

//BEG - Global variables
	PARAMS			g_AppParam;		//  

	LPDIRECT3D9             g_pD3D       = NULL;	// Used to create the D3DDevice
	LPDIRECT3DDEVICE9       g_pd3dDevice = NULL;	// Our rendering device
	LPDIRECT3DTEXTURE9      g_pHMTexture = NULL;	// HeightMap texture
	LPDIRECT3DTEXTURE9      g_pTMTexture = NULL;	// TreeMap texture
	LPDIRECT3DTEXTURE9      g_pOutTexture = NULL;	// Out HorizonMap texture

//END - Global variables


//BEG - functions declarations

	//Initializes Direct3D
	HRESULT InitD3D( HWND hWnd );
	//Load HM & TM Textures, creating Out texture
	HRESULT InitTextures();
	//Update HeightMap, by info of a TreeMap
	HRESULT UpdateHeightMap();
	//Calculate pow2
	float sqr(float num);
	//Generate shadow horizon map
	HRESULT GenerateHorizoneMap();
	//It will make a shadows under trees
	HRESULT AddShadowUnderTrees();
	HRESULT SaveHorizonMap();
	//Before exit
	void FINISH (HRESULT hr);

//END - functions declarations


//--------------------------------------------------------------------------

//Initializes Direct3D
HRESULT InitD3D( HWND hWnd )
{
	// Create the D3D object.
	if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
		return E_FAIL;

	// Set up the structure used to create the D3DDevice. Since we are now
	// using more complex geometry, we will create a device with a zbuffer.
	D3DPRESENT_PARAMETERS d3dpp;
	ZeroMemory( &d3dpp, sizeof(d3dpp) );
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	d3dpp.EnableAutoDepthStencil = TRUE;
	d3dpp.AutoDepthStencilFormat = D3DFMT_D16;

	// Create the D3DDevice
	if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING,
		&d3dpp, &g_pd3dDevice ) ) )
	{
		return E_FAIL;
	}

	return S_OK;
}

//Load Textures
HRESULT InitTextures()
{
	HRESULT hr;
	printf ("Loading textures");
	//Loading HeightMap
	hr = D3DXCreateTextureFromFileEx(g_pd3dDevice,					//D3DDevice
		g_AppParam.chHMfname,			//File name
		g_AppParam.iOutRezolution,		//Width
		g_AppParam.iOutRezolution,		//Height
		0,								//MipLevels
		0,								//DWORD Usage,
		D3DFMT_A32B32G32R32F,//D3DFMT_R32F,	//D3DFORMAT Format,
		D3DPOOL_SYSTEMMEM,				//D3DPOOL Pool,
		D3DX_FILTER_TRIANGLE,			//Filter
		D3DX_FILTER_TRIANGLE,			//Mip Filter
		0,								//D3DCOLOR ColorKey,
		NULL,							//D3DXIMAGE_INFO *pSrcInfo,
		NULL,							//PALETTEENTRY *pPalette,
		&g_pHMTexture					//Texture pointer
		);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t load height map.\n");
		return hr;
	}
	printf ("....");
	//Loading TreeMap
	hr = D3DXCreateTextureFromFileEx(g_pd3dDevice,					//D3DDevice
		g_AppParam.chTMfname,			//File name
		g_AppParam.iOutRezolution,		//Width
		g_AppParam.iOutRezolution,		//Height
		0,								//MipLevels
		0,								//DWORD Usage,
		D3DFMT_A32B32G32R32F,//D3DFMT_R32F,	//D3DFORMAT Format,
		D3DPOOL_SYSTEMMEM,				//D3DPOOL Pool,
		D3DX_FILTER_TRIANGLE,			//Filter
		D3DX_FILTER_TRIANGLE,			//Mip Filter
		0,								//D3DCOLOR ColorKey,
		NULL,							//D3DXIMAGE_INFO *pSrcInfo,
		NULL,							//PALETTEENTRY *pPalette,
		&g_pTMTexture					//Texture pointer
		);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t load tree map.\n");
		return hr;
	}
	printf ("....");
	//Creating Out HorizonMap Texture
	hr = D3DXCreateTexture(g_pd3dDevice,					//D3DDevice
		g_AppParam.iOutRezolution,		//Width
		g_AppParam.iOutRezolution,		//Height
		0,								//MipLevels
		0,								//DWORD Usage,
		D3DFMT_A32B32G32R32F,//D3DFMT_R32F,	//D3DFORMAT Format,
		D3DPOOL_SYSTEMMEM,				//D3DPOOL Pool
		&g_pOutTexture
		);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t create new texture for horizon map.\n");
		return hr;
	}
	printf ("..");
	printf("OK\n");
	return S_OK;
}

//Update HeightMap, by info of a TreeMap
//        ,       ,
//      
HRESULT UpdateHeightMap()
{
	HRESULT hr;
	printf("Mixing height map & tree map");

	// 
	D3DSURFACE_DESC HMDesk;
	ZeroMemory(&HMDesk,sizeof(D3DSURFACE_DESC));
	g_pHMTexture->GetLevelDesc(0,&HMDesk);
	D3DSURFACE_DESC TMDesk;
	ZeroMemory(&TMDesk,sizeof(D3DSURFACE_DESC));
	g_pTMTexture->GetLevelDesc(0,&TMDesk);
	
	// 
	D3DLOCKED_RECT HMLockedRect;
	D3DLOCKED_RECT TMLockedRect;
	hr = g_pHMTexture->LockRect(0,&HMLockedRect, NULL, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock HeightMap texture!\n");
		return E_FAIL;
	}
	hr = g_pTMTexture->LockRect(0,&TMLockedRect, NULL, D3DLOCK_READONLY);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock TreeMap texture!\n");
		return E_FAIL;
	}
	printf (".");

	//  
	D3DXVECTOR4* pHM = (D3DXVECTOR4*)HMLockedRect.pBits;
	D3DXVECTOR4* pTM = (D3DXVECTOR4*)TMLockedRect.pBits;
	
	//      ,    M    TREE_HEIGHT
	for (UINT i=0;i<=TMDesk.Height;i++)
		for (UINT j=0;j<=TMDesk.Width;j++)
		{			
			if (pTM[i*TMDesk.Width+j].x>0.1/*128.0f/256.0f*/)
			{
				pHM[i*TMDesk.Width+j].x+=TREE_HEIGHT/g_AppParam.fHeight;
				pHM[i*TMDesk.Width+j].y+=TREE_HEIGHT/g_AppParam.fHeight;
				pHM[i*TMDesk.Width+j].z+=TREE_HEIGHT/g_AppParam.fHeight;
			}
		}
	printf ("......");
	// 
	g_pHMTexture->UnlockRect(0);
	g_pTMTexture->UnlockRect(0);
	//Test HM texture to file D3DXSaveTextureToFile(L"testHM.dds", D3DXIFF_DDS, g_pHMTexture, NULL);
	printf ("..");
	printf("OK\n");
	return S_OK;
}

//Calculate pow2
float sqr(float num)
{
	return pow(num,2);
}

//   ( )
HRESULT GenerateHorizoneMap()
{
	HRESULT hr;
	printf ("Generating Horizon maps");

	// 
	D3DSURFACE_DESC HMDesk;
	ZeroMemory(&HMDesk,sizeof(D3DSURFACE_DESC));
	g_pHMTexture->GetLevelDesc(0,&HMDesk);
	D3DSURFACE_DESC OutDesk;
	ZeroMemory(&OutDesk,sizeof(D3DSURFACE_DESC));
	//g_pTMTexture->GetLevelDesc(0,&OutDesk);
	g_pOutTexture->GetLevelDesc(0,&OutDesk);

	// 
	D3DLOCKED_RECT HMLockedRect;
	D3DLOCKED_RECT OutLockedRect;
	hr = g_pHMTexture->LockRect(0,&HMLockedRect, NULL, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock HeightMap texture!\n");
		return E_FAIL;
	}
	hr = g_pOutTexture->LockRect(0,&OutLockedRect, NULL, D3DLOCK_READONLY);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock OutTexture!\n");
		return E_FAIL;
	}
	printf (".");

	//  
	D3DXVECTOR4* pHM =  (D3DXVECTOR4*)HMLockedRect.pBits;
	D3DXVECTOR4* pOut = (D3DXVECTOR4*)OutLockedRect.pBits;

	// 
	float fPixelSize = g_AppParam.fTerrainSize/(float)HMDesk.Width; //  
	float fHeightSize= g_AppParam.fHeight;//   

	int iCount = 0; // ,    
	DWORD dwStart = GetTickCount(); //   
	//   :
	int xp=1;
	//   : +x
	for (UINT i=0;i<=HMDesk.Height;i++)
	{
		for (UINT j=0;j<=HMDesk.Width;j++)
		{

			//   t (xt,yt,zt)  
			D3DXVECTOR3 Tt = D3DXVECTOR3(float(i), //Xt
				float(j), //Yt
				pHM[i*HMDesk.Width+j].x //Zt (height)
				);
			//  
			Tt = D3DXVECTOR3(Tt.x*fPixelSize,Tt.y*fPixelSize,Tt.z*fHeightSize + g_AppParam.fHeightZeroPoint);
			D3DXVECTOR3 Hmax;
			//   ( )
			float Xmax = 900000.0f;//HMDesk.Height*fPixelSize;	 //The distance to sun (meters) :-)
			//        (xp, yp) 
			UINT x = i; //    
			float min_cos_alfa = 1.0f;
			//  
			while (x<HMDesk.Height)
			{
				// 
				x=x+xp;
				Hmax.x = float(x);
				Hmax.y = float(j);
				Hmax.z = pHM[x*HMDesk.Width+j].x;
				//  
				Hmax = D3DXVECTOR3(Hmax.x*fPixelSize,Hmax.y*fPixelSize,Hmax.z*fHeightSize + g_AppParam.fHeightZeroPoint);
				float cos_alfa;
				if (Hmax.z<Tt.z)
				{
					cos_alfa = 0.9999999f;
				} 
				else
				{
					float cos_t = (Hmax.x - Tt.x)/sqrt(sqr(Hmax.x - Tt.x)+sqr(Hmax.z - Tt.z));
					float Z = sqrt(sqr((Xmax-Tt.x)/cos_t)-sqr(Xmax-Tt.x));
					float Hsun = Tt.z + Z;
					cos_alfa = (Xmax/2.0f)/sqrt(sqr(Xmax/2.0f)+sqr(Hsun));
				}

				//  ,  
				if (min_cos_alfa>cos_alfa)
				{
					min_cos_alfa=cos_alfa;
				}
			}
			// cos_alfa -          .
			pOut[i*OutDesk.Width+j].x = min_cos_alfa;
		}
		//  , ,   ,   
		iCount++;
		if (iCount>100)
		{
			printf (".");
			iCount=0;
		}
	}

	xp=-1;
	//   : -x
	for (INT i=HMDesk.Height;i>=0;i--)
	{
		for (INT j=HMDesk.Width;j>=0;j--)
		{

			//   t (xt,yt,zt)  
			D3DXVECTOR3 Tt = D3DXVECTOR3(float(i), //Xt
				float(j), //Yt
				pHM[i*HMDesk.Width+j].x //Zt (height)
				);
			//  
			Tt = D3DXVECTOR3(Tt.x*fPixelSize,Tt.y*fPixelSize,Tt.z*fHeightSize +  g_AppParam.fHeightZeroPoint);
			D3DXVECTOR3 Hmax;
			//   ( )
			float Xmax = 900000.0f;//0; //The distance to sun (meters) :-)
			//float Co = (HMDesk.Height/2.0f)*fPixelSize; //Center of map (meters)
			float Co = Xmax/2.0f;
			//        (xp, yp) 
			INT x = i; //    
			float min_cos_alfa = 1.0f;
			//  
			while (x>0)
			{
				// 
				x=x+xp;
				Hmax.x = float(x);
				Hmax.y = float(j);
				Hmax.z = pHM[x*HMDesk.Width+j].x;
				//  
				Hmax = D3DXVECTOR3(Hmax.x*fPixelSize,Hmax.y*fPixelSize,Hmax.z*fHeightSize +  g_AppParam.fHeightZeroPoint);
				float cos_alfa;
				if (Hmax.z<Tt.z)
				{
					cos_alfa = 0.9999999f;
				} 
				else
				{
					float cos_t = (Hmax.x - Tt.x)/sqrt(sqr(Hmax.x - Tt.x)+sqr(Hmax.z - Tt.z));
					float Z = sqrt(sqr((Xmax-Tt.x)/cos_t)-sqr(Xmax-Tt.x));
					float Hsun = Tt.z + Z;
					cos_alfa = (Co)/sqrt(sqr(Co)+sqr(Hsun));
				}

				//  ,  
				if (min_cos_alfa>cos_alfa)
				{
					min_cos_alfa=cos_alfa;
				}
			}
			// cos_alfa -          .
			pOut[i*OutDesk.Width+j].y = min_cos_alfa;
		}
		//  , ,   ,   
		iCount++;
		if (iCount>100)
		{
			printf (".");
			iCount=0;
		}
	}

	// 
	g_pHMTexture->UnlockRect(0);
	g_pOutTexture->UnlockRect(0);
	printf (".");
	printf("OK\n");
	printf("Time spend for operation: %f sec.\n", (GetTickCount() - dwStart)/1000.0f);
	return S_OK;
}


//   ( )  2  (  )
HRESULT GenerateHorizoneMap2()
{
	HRESULT hr;
	printf ("Generating Horizon maps optimized");
	
	// 
	D3DSURFACE_DESC HMDesk;
	ZeroMemory(&HMDesk,sizeof(D3DSURFACE_DESC));
	g_pHMTexture->GetLevelDesc(0,&HMDesk);
	D3DSURFACE_DESC OutDesk;
	ZeroMemory(&OutDesk,sizeof(D3DSURFACE_DESC));
	//g_pTMTexture->GetLevelDesc(0,&OutDesk);
	g_pOutTexture->GetLevelDesc(0,&OutDesk);

	// 
	D3DLOCKED_RECT HMLockedRect;
	D3DLOCKED_RECT OutLockedRect;
	hr = g_pHMTexture->LockRect(0,&HMLockedRect, NULL, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock HeightMap texture!\n");
		return E_FAIL;
	}
	hr = g_pOutTexture->LockRect(0,&OutLockedRect, NULL, D3DLOCK_READONLY);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock OutTexture!\n");
		return E_FAIL;
	}
	printf (".");

	//  
	D3DXVECTOR4* pHM =  (D3DXVECTOR4*)HMLockedRect.pBits;
	D3DXVECTOR4* pOut = (D3DXVECTOR4*)OutLockedRect.pBits;

	// 
	float fPixelSize = g_AppParam.fTerrainSize/(float)HMDesk.Width; //  
	float fHeightSize= g_AppParam.fHeight;//   

	int iCount = 0; // ,    
	DWORD dwStart = GetTickCount(); //   
	//   :
	float cos_alfa = 0;
	//int xp=1;
	float Xc = g_AppParam.fTerrainSize/2.0f;
	float Xs = 900000.0f;//g_AppParam.fTerrainSize;//*2.0f;//
	//   : +x
	for (UINT i=0;i<=HMDesk.Height;i++)
	{
		for (UINT j=0;j<=HMDesk.Width;j++)
		{
			//     
			float Xt = i*fPixelSize; //        
			float Ht = pHM[i*HMDesk.Width+j].x; //   
			//  
			Ht = Ht*fHeightSize + g_AppParam.fHeightZeroPoint;
			//,   ,    
			float Xi;
			float Hi;
			float HiMax=0.0f; float XiMax=0.0f; int iMax=0; //,    ()
			float Hx = 0.0f;
			bool  bIsShaded = FALSE; //   
			//        (xp)
			for (UINT ii=i+1;ii<=HMDesk.Height;ii++)
			{
				Xi=ii*fPixelSize; //     
				Hi=pHM[ii*HMDesk.Width+j].x*fHeightSize + g_AppParam.fHeightZeroPoint;//   
				if (Hi<=Ht) //   .  .     
				{
					cos_alfa = 0.999f;
					continue;
					//   
				}
				else 
				{
					float Hxi = (Hi - Ht)/(Xi - Xt);
					if (Hxi>Hx) //    -   ,   
					{
						Hx = Hxi;
						iMax = ii;
						HiMax = Hi;
						XiMax = Xi;
						bIsShaded = TRUE;
					}
				}
			}
			//  ,     ,     ,   
			if (bIsShaded)
			{
				float Hs = ((HiMax-Ht)/(XiMax-Xt))*(Xs - Xt);
				float tg_alfa = (Hs+Ht)/(Xs-Xc);
				cos_alfa = 1.0f/sqrt(1+sqr(tg_alfa));
			}
			// cos_alfa -          t.
			pOut[i*OutDesk.Width+j].x = cos_alfa;
		}
		//  , ,   ,   
		iCount++;
		if (iCount>100)
		{
			printf (".");
			iCount=0;
		}
	}

	//   : -x
	for (INT i=HMDesk.Height;i>=0;i--)
	{
		for (INT j=HMDesk.Width;j>=0;j--)
		{//     
			float Xt = i*fPixelSize; //        
			float Ht = pHM[i*HMDesk.Width+j].x; //   
			//  
			Ht = Ht*fHeightSize + g_AppParam.fHeightZeroPoint;
			//,   ,    
			float Xi;
			float Hi;
			float HiMax=0.0f; float XiMax=0.0f; int iMax=0; //,    ()
			float Hx = 0.0f;
			bool  bIsShaded = FALSE; //   
			//        (xp)
			for (INT ii=i-1;ii>=0;ii--)
			{
				Xi=ii*fPixelSize; //     
				Hi=pHM[ii*HMDesk.Width+j].x*fHeightSize + g_AppParam.fHeightZeroPoint;//   
				if (Hi<=Ht) //   .  .     
				{
					cos_alfa = 0.999f;
					continue;
					//   
				}
				else 
				{
					float Hxi = (Hi - Ht)/(Xt - Xi);
					if (Hxi>Hx) //    -   ,   
					{
						Hx = Hxi;
						iMax = ii;
						HiMax = Hi;
						XiMax = Xi;
						bIsShaded = TRUE;
					}
				}
			}
			//  ,     ,     ,   
			if (bIsShaded)
			{
				float Hs = ((HiMax-Ht)/(Xt-XiMax))*(Xs + Xt);
				cos_alfa = 1.0f/sqrt(1+sqr((Hs+Ht)/(Xs + Xc)));
			}
			// cos_alfa -          t.
			pOut[i*OutDesk.Width+j].y = cos_alfa;
		}
		//  , ,   ,   
		iCount++;
		if (iCount>100)
		{
			printf (".");
			iCount=0;
		}
	}

	// 
	g_pHMTexture->UnlockRect(0);
	g_pOutTexture->UnlockRect(0);
	printf (".");
	printf("OK\n");
	printf("Time spend for operation: %f sec.\n", (GetTickCount() - dwStart)/1000.0f);
	return S_OK;
}

//It will make a shadows under trees
HRESULT AddShadowUnderTrees()
{
	HRESULT hr;
	printf("Mixing horizon map & tree map");

	// 
	D3DSURFACE_DESC HMDesk;
	ZeroMemory(&HMDesk,sizeof(D3DSURFACE_DESC));
	g_pOutTexture->GetLevelDesc(0,&HMDesk); //!!! <-  HMDesk  ,    !!!
	D3DSURFACE_DESC TMDesk;
	ZeroMemory(&TMDesk,sizeof(D3DSURFACE_DESC));
	g_pTMTexture->GetLevelDesc(0,&TMDesk);

	// 
	D3DLOCKED_RECT HMLockedRect;			//!!! <-  HMLockedRect  ,    !!!
	D3DLOCKED_RECT TMLockedRect;
	hr = g_pOutTexture->LockRect(0,&HMLockedRect, NULL, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock HeightMap texture!\n");
		return E_FAIL;
	}
	hr = g_pTMTexture->LockRect(0,&TMLockedRect, NULL, D3DLOCK_READONLY);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t lock TreeMap texture!\n");
		return E_FAIL;
	}
	printf (".");

	//  
	D3DXVECTOR4* pHM = (D3DXVECTOR4*)HMLockedRect.pBits; //!!! <-  HM  ,    !!!
	D3DXVECTOR4* pTM = (D3DXVECTOR4*)TMLockedRect.pBits;

	//      ,       :-)
	for (UINT i=0;i<=TMDesk.Height;i++)
		for (UINT j=0;j<=TMDesk.Width;j++)
		{			
			if (pTM[i*TMDesk.Width+j].x>0.1/*128.0f/256.0f*/)
			{
				pHM[i*TMDesk.Width+j].x=0.001f;
				pHM[i*TMDesk.Width+j].y=0.001f;
			}
		}
		printf ("......");
		// 
		g_pOutTexture->UnlockRect(0);
		g_pTMTexture->UnlockRect(0);
		//Test HM texture to file D3DXSaveTextureToFile(L"testHM.dds", D3DXIFF_DDS, g_pHMTexture, NULL);
		printf ("..");
		printf("OK\n");
		return S_OK;
}

//Writing DDS file with HorizonMap
HRESULT SaveHorizonMap()
{
	HRESULT hr;
	printf ("Writing HorizonMap file");

	//       
	hr = D3DXSaveTextureToFile(g_AppParam.chOutfname, D3DXIFF_DDS, g_pOutTexture, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t save HorizonMap texture!\n");
		return E_FAIL;
	}
	printf ("...");
	//        (winter)

	// Search backward for '.'
	/*
	int		result = (int)(pdest - g_AppParam.chOutfname);
	wchar_t wchTemp[MAX_PATH];*/
	//Rename to winter texture
	wchar_t drive	[_MAX_DRIVE];
	wchar_t	dir		[_MAX_DIR];
	wchar_t fname	[_MAX_FNAME];
	wchar_t	ext		[_MAX_EXT],extWinter[_MAX_EXT];
	wchar_t FullName[MAX_PATH];
	_wsplitpath ( g_AppParam.chOutfname, drive, dir, fname, ext );
	wcscpy(extWinter,L".winter");
	wcscat(extWinter,ext);
	_wmakepath (FullName, drive, dir, fname, extWinter );
	
	hr = D3DXSaveTextureToFile(FullName, D3DXIFF_DDS, g_pOutTexture, NULL);
	if (hr!=S_OK)
	{
		printf("ERROR:Can`t save HorizonMap texture!\n");
		return E_FAIL;
	}
	printf ("...");
	printf("OK\n");
	return S_OK;
}

//Main function
int _tmain(int argc, _TCHAR* argv[])
{
	HRESULT hr;
	printf("\nHorizonator v 1.1.\nAutor: Morock, 2007\n\n");
	if (!g_AppParam.InitParams(argc)) {FINISH(E_FAIL);return 0;}

	g_AppParam.UseParams();
	
	if(g_AppParam.TestParams())
	{
	//Main code here
		HWND hWnd = GetDesktopWindow();
		//init D3DX Device
		hr = InitD3D(hWnd);
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t create D3DX device.\n");
			FINISH(hr);
			return 1;
		}
		//Load & Init textures
		hr = InitTextures();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t init textures.\n");
			FINISH(hr);
			return 2;
		}
		//Split heights from Height Map & heights of trees
		hr = UpdateHeightMap();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t split HeightMap with a TreeMap.\n");
			FINISH(hr);
			return 3;
		}
		//Basing on a updated HeightMap generate HorizonMap
		//hr = GenerateHorizoneMap();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t generate HorizoneMap.\n");
			FINISH(hr);
			return 4;
		}
		hr = GenerateHorizoneMap2();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t generate HorizoneMap.\n");
			FINISH(hr);
			return 4;
		}
		//And now we`ll correct artifact of a light under tree (will make a shadows under trees)
		hr = AddShadowUnderTrees();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t split HorizonMap with a TreeMap.\n");
			FINISH(hr);
			return 5;
		}
		//Writing HorizonMap to DDS file
		hr = SaveHorizonMap();
		if (hr!=S_OK)
		{
			printf("ERROR:Can`t save file.\n");
			FINISH(hr);
			return 6;
		}
	}
	
	//End program
	FINISH(S_OK);
	return 0;
}


//Before exit
void FINISH(HRESULT hr)
{
	if (g_pOutTexture)
	{
		g_pOutTexture->Release();
	}
	if (g_pTMTexture)
	{
		g_pTMTexture->Release();
	}
	if (g_pHMTexture)
	{
		g_pHMTexture->Release();
	}
	if (g_pd3dDevice)
	{
		g_pd3dDevice->Release();
	}
	if (g_pD3D)
	{
		g_pD3D->Release();
	}
	if (hr!=S_OK)
	{
		MessageBeep(MB_ICONEXCLAMATION);
		const WCHAR  * chErrorDesc;
		chErrorDesc = DXGetErrorDescription9(hr);
		printf("Reason: %S.\nApplication work finished with error. Press any key for exit...", chErrorDesc);
		char ch = _getch();
	}
	else
	{
		printf("Application work successfully finished. Good day!");
		//_getch();
	}
}
