
//////////////////////////////////////////
// UnTweakables, assigned automatically //
//////////////////////////////////////////

float4x4 WvpXf : WorldViewProjection < string UIWidget="None"; >;

///////////////////////////////////////////////
//  Tweakables Appear in the Properties Pane //
///////////////////////////////////////////////

float SunAngle   <
    string UIWidget = "slider";
    float UIMin = 0.0;
    float UIMax = 180.0;
    float UIStep = 0.1;
    string UIName = "Sun Angle";
> = 90.0;

////////////////////////////////////
//  Texture  ///////////////////////
////////////////////////////////////

texture diffuseTexture : Diffuse;
texture horizonTexture;

sampler DSampler = sampler_state 
{
    texture = <diffuseTexture>;
    AddressU  = CLAMP;        
    AddressV  = CLAMP;
    AddressW  = CLAMP;
    MIPFILTER = LINEAR;
    MINFILTER = LINEAR;
    MAGFILTER = LINEAR;
    /*
    MagFilter = POINT; 
 	MinFilter = POINT;
 	MipFilter = NONE;
    */
};

sampler HSampler = sampler_state 
{
    texture = <horizonTexture>;
    AddressU  = CLAMP;        
    AddressV  = CLAMP;
    AddressW  = CLAMP;
    MIPFILTER = ANISOTROPIC;
    MINFILTER = ANISOTROPIC;
    MAGFILTER = ANISOTROPIC;
    MaxAnisotropy = 20;
};

////////////////////////////////////
//  Structures Used ////////////////
////////////////////////////////////

struct vertexInput {
    float3 position				: POSITION;
    float3 normal				: NORMAL;
    float4 texCoordDiffuse		: TEXCOORD0;
};

struct vertexOutput {
    float4 hPosition		: POSITION;
    float4 texCoordDiffuse	: TEXCOORD0;
    float4 shad				: COLOR;
};

////////////////////////////////////
// Vertex Shader ///////////////////
////////////////////////////////////

vertexOutput VS_TransformAndTexture(vertexInput IN) 
{
    vertexOutput OUT = (vertexOutput)0;
    float4 Po = float4(IN.position.xyz,1.0); // object-coordinates position
    OUT.hPosition = mul(Po,WvpXf);
    OUT.texCoordDiffuse = IN.texCoordDiffuse;	
    return OUT;
}

////////////////////////////////////
// Pixel Shader ////////////////////
////////////////////////////////////
//HorizonMaping with shadow blendings

float4 PS_Horizon( vertexOutput IN): COLOR
{
	float4 diffuseTexture = tex2D(DSampler, IN.texCoordDiffuse );
	float fAdder  = 0.0006;
	int Quality=4;
 	float2 cosAngle[9];
 	cosAngle[0] = tex2D(HSampler, IN.texCoordDiffuse);
 	cosAngle[1] = tex2D(HSampler, float2(IN.texCoordDiffuse.x+fAdder,IN.texCoordDiffuse.y+fAdder)).xy;
 	cosAngle[2] = tex2D(HSampler, float2(IN.texCoordDiffuse.x-fAdder,IN.texCoordDiffuse.y-fAdder)).xy;
    cosAngle[3] = tex2D(HSampler, float2(IN.texCoordDiffuse.x+fAdder,IN.texCoordDiffuse.y-fAdder)).xy;
    cosAngle[4] = tex2D(HSampler, float2(IN.texCoordDiffuse.x-fAdder,IN.texCoordDiffuse.y+fAdder)).xy;
    cosAngle[5] = tex2D(HSampler, float2(IN.texCoordDiffuse.x+fAdder,IN.texCoordDiffuse.y)).xy;
 	cosAngle[6] = tex2D(HSampler, float2(IN.texCoordDiffuse.x-fAdder,IN.texCoordDiffuse.y)).xy;
    cosAngle[7] = tex2D(HSampler, float2(IN.texCoordDiffuse.x,IN.texCoordDiffuse.y-fAdder)).xy;
    cosAngle[8] = tex2D(HSampler, float2(IN.texCoordDiffuse.x,IN.texCoordDiffuse.y+fAdder)).xy;
 	float4 rez = 0;
	for (int i=0; i<Quality; i++)
	{
	if (SunAngle<90)

		//soft shadows
		//return diffuseTexture*(1+saturate(SunAngle - degrees(acos(cosAngle.r))));
		rez+=(0.5f+saturate(SunAngle - degrees(acos(cosAngle[i].x))));
	else

		//soft shadows
		//return diffuseTexture*(1+saturate(180 - SunAngle - degrees(acos(cosAngle.g))));
		rez+=(0.5f+saturate(180-SunAngle - degrees(acos(cosAngle[i].y))));
	}
	return rez*diffuseTexture/Quality;
}

technique textured
{
    pass p0 
    {		
		VertexShader = compile vs_2_0 VS_TransformAndTexture();
		PixelShader  = compile ps_2_a PS_Horizon();
    }
     
}

